/* threads v1.0, functions for cooperative multithreading
 * Copyright 2002, 2003 Chris Williams <cow3.14@juno.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Started       2002 Jan 09
 * Last modified 2003 Jun 24
 */

#include "threads.h"

list_t threads;
thread_t main_thread;
int num_running_threads;

/* push_all and pop_all are assembly macros */
inline void push_all() {}
inline void pop_all() {}

/* halt is the HALT instruction. */
inline void halt() {}

void threads_init(void)
{
	threads.head = threads.tail = NULL;
	main_thread.sp = main_thread.stack_top = sp;
	main_thread.state = STATE_RUNNING;
	list_append_node(&threads, (node_t *)&main_thread);
	num_running_threads = 1;
}

int thread_create(thread_t *t, void *(*code)(void), void *stack_top)
{
	if (t == NULL)
		return -1;
	t->code = code;
	t->stack_top = stack_top;
	t->state = STATE_STARTED;
	list_append_node(&threads, &t->node);
	/*thread_cooperate();*/
	return 0;
}

int thread_setstate(thread_t *t, enum thread_state state)
{
	if (state != STATE_RUNNING && state != STATE_SUSPENDED)
		return -1;
	
	if (t->state == state)
		return 0;
	
	if (state == STATE_RUNNING)
		++num_running_threads;
	else if (t->state == STATE_RUNNING)
		--num_running_threads;
	
	t->state = state;
	return 0;
}

int thread_cancel(thread_t *t)
{
	if (t == NULL)
		return -1;
	
	/* exit immediately if the main thread is cancelled */
	if (t == &main_thread) {
		sp = main_thread.stack_top;
		return 0;
	}
	
	if (t->state == STATE_RUNNING)
		--num_running_threads;
	t->state = STATE_CANCELLED;
	thread_cooperate();
	/* if t is the current, thread_cooperate() will not return here */
	return 0;
}

void thread_suspend(void)
{
	enum thread_state *s = &((thread_t *)threads.head)->state;
	if (*s == STATE_SUSPENDED)
		thread_cooperate();
	else {
		*s = STATE_SUSPENDED;
		--num_running_threads;
		thread_cooperate();
		++num_running_threads;
		*s = STATE_RUNNING;
	}
}

void thread_cooperate(void)
{
	push_all(); /* save all registers */
	
	threads.tail = threads.head;
	((thread_t *)threads.head)->sp = sp;
	
	threads.head = threads.head->next; /* go to next thread in the list */
	
	while (1) {
		thread_t *t = (thread_t *)threads.head;
		switch (t->state) {
running:	case STATE_RUNNING:
			sp = t->sp; /* go to thread's stack space */
			pop_all(); /* restore all registers */
			return; /* return to a running thread */
		case STATE_SUSPENDED:
			if (t == &main_thread
			  && num_running_threads == 0)
				halt();
			/* fall through */
			goto running;
		case STATE_CANCELLED:
			thread_exit(0); /* this does not return here */
		case STATE_STARTED:
			sp = t->stack_top; /* go to thread's stack space */
			
			++num_running_threads;
			t->state = STATE_RUNNING;
			t->retval = t->code();
			/* every thread (except the main thread) "returns" here
			 * when it ends, either by returning, by calling
			 * thread_exit, or through thread_cancel. */
			if (t->state == STATE_RUNNING)
				--num_running_threads;
			if (t->state != STATE_CANCELLED)
				t->state = STATE_EXITED;
		/* fall through */
		default:
			list_remove_head(&threads);
			/* We don't want to keep a thread who has stopped
			 * (STATE_CANCELLED or STATE_EXITED) or whose state
			 * is invalid. */
		}
	}
}

void *thread_exit(void *retval)
{
	sp = ((thread_t *)threads.head)->stack_top;
	return retval;
}

void *thread_join(thread_t *t)
{
	while (t->state != STATE_EXITED && t->state != STATE_CANCELLED)
		thread_cooperate();
	
	return t->retval;
}
